
#include <string>
#include <fstream>
#include <iostream>

#include <Glob.hh>
#include <support.hh>
#include <os.hh>

#include <inttypes.h>

#include <input-match.hh>

static char *bits_envvar[] = { "EV", "KEY", "REL", "ABS", "MSC", "LED", "SND", "FF", NULL };

enum { // same order as BITS_* in input-match.hh !!!
  MATCH_BUS     = 0x0001,
  MATCH_VENDOR  = 0x0002,
  MATCH_PRODUCT = 0x0004,
  MATCH_VERSION = 0x0008,
  MATCH_EVBIT   = 0x0010,
  MATCH_KEYBIT  = 0x0020,
  MATCH_RELBIT  = 0x0040,
  MATCH_ABSBIT  = 0x0080,
  MATCH_MSCBIT  = 0x0100,
  MATCH_LEDBIT  = 0x0200,
  MATCH_SNDBIT  = 0x0400,
  MATCH_FFBIT   = 0x0800,
};

bool input_match_bits(const std::vector<uint32_t> &mod_bits, const std::vector<uint32_t> &dev_bits) {
  /* NOTE this tries to follow input.agent from hotplut-2004-09-23, 
     but I think it's wrong (watch how mword and dword is
     actually never updated there) ... so this is a guess */
  
  int cnt = 1;
  while (true) {
    uint32_t mword = (cnt > mod_bits.size()) ? 0 : mod_bits[mod_bits.size()-cnt];;
    uint32_t dword = (cnt > dev_bits.size()) ? 0 : dev_bits[dev_bits.size()-cnt];
    
    if (mword & dword != mword) {
      return false;
    }
    
    cnt++;
    if ((cnt > mod_bits.size()) && (cnt > dev_bits.size())) {
      break;
    }
    else if ((cnt > mod_bits.size()) || (cnt > dev_bits.size())) {
      return false;
    }
  }

  return false; /* FIXME no match at all for now... bit matching has to be revised */
}

bool input_module::operator== (const input_module& dev) {
  if ((matchbits & MATCH_BUS) && (bustype != dev.bustype)) return false;
  if ((matchbits & MATCH_VENDOR) && (vendor != dev.vendor)) return false;
  if ((matchbits & MATCH_PRODUCT) && (product != dev.product)) return false;

  if ((matchbits & MATCH_VERSION) && (version < dev.version)) return false;

  for(int i=0; i<BITS_LAST; i++) {
    if (matchbits & (MATCH_EVBIT << i)) {
      if (!input_match_bits(bits[i], dev.bits[i])) {
	return false;
      }
    }
  }

  return true;
}

void input_module::read_from_sysfs (const std::string& file) {
  /* FIXME */
}

void input_module::read_from_proc (const std::string& file) {
  /* FIXME */
}

void input_module::read_from_env () {
  const char *str;
  if ((str = getenv("PRODUCT"))) {
    sscanf(str, "%x/%x/%x/%x", &bustype, &vendor, &product, &version);
  }
  else {
    bustype = vendor = product = version = 0;
  }

  name = getenv("NAME");
  phys = getenv("PHYS");

  matchbits = 0;

  for(int i=0; i<BITS_LAST; i++) {
    const char *ptr = getenv(bits_envvar[i]);

    /* clear and check */
    bits[i] = std::vector<uint32_t>();

    /* parse array */
    while (ptr && *ptr) {
      char *eptr;
      uint32_t val = strtol(ptr, &eptr, 16);

      if (eptr == ptr) break;

      bits[i].push_back(val);
      ptr = eptr;
      while ((*ptr == ':') || (*ptr == ' ') || (*ptr == '\t')) ptr++; /* FIXME <- really needed? */
    }
  }
}

std::ostream& operator<< (std::ostream& os, const input_module& m) {
  os.setf (std::ios::hex, std::ios::basefield);

  os << m.name << " (" << m.phys << ") "
     << m.bustype << "/" << m.vendor << "/" << m.product << "/" << m.version
     << " [" << m.matchbits << "]";

  for(int i=0, j=0; i<input_module::BITS_LAST; i++) {
    if (!m.bits[i].empty()) {
      os << " " << bits_envvar[i] << "=";
      for(; j<m.bits[i].size()-1;j++) os << m.bits[i][j] << ":";
      os << m.bits[i][j];
    }
  }

  os << " INFO=" << m.driver_info;

  if (!m.module.empty())
    os << ": Module: " << m.module;
  
  return os;
}

// ...

std::vector<input_module> input_subsystem::modules;

void input_subsystem::hotplug ()
{
  parse_modules();
  input_module m;
  m.read_from_env ();
  match_modules (m, &input_subsystem::hotplug);
}

void input_subsystem::hotplug (const input_module& m)
{
  if (m.module.size()) {
    if (m.module == "evbug") {
      std::cout << "IGNORING modprobe " << m.module << std::endl;
      return;
    }

    modprobe (m.module);
  }
}

void input_subsystem::hotplug_all ()
{
  parse_modules();
  for_each (&input_subsystem::hotplug);
}

void input_subsystem::list ()
{
  std::cout << "INPUT Devices and matched modules:" << std::endl;
  parse_modules();
  
  for_each (&input_subsystem::list);
  std::cout << std::endl;
}

void input_subsystem::list (const input_module& m)
{
  std::cout << m << std::endl;
}

void input_subsystem::match_modules (input_module& m,
				   void(input_subsystem::*op)(const input_module&))
{
  std::vector<input_module>::iterator module = modules.begin();
  while (module != modules.end()) {
    module = std::find (module, modules.end(), m);
    if (module != modules.end()) {
      m.module = module->module;
        (this->*op)(m);
	++module;
    }
  }
}

void input_subsystem::for_each (void(input_subsystem::*op)(const input_module&))
{
  // FIXME is there a input device list somewhere really?
  // /sys/class/input/*/dev only spits out device major:minor number
#if 0
  Utility::Glob input_devices ("/sys/class/input/*/dev");
  for (Utility::Glob::iterator it = input_devices.begin();
       it != input_devices.end(); ++it)
    {
      input_module m;
      m.read_from_sysfs (*it);
      match_modules (m, op);
    }
  if (input_devices)
    return;

  Utility::Glob input_devices2 ("/proc/bus/input/*/??.?");
  for (Utility::Glob::iterator it = input_devices2.begin();
       it != input_devices2.end(); ++it)
  {
    /* FIXME 
       input_module m;
       m.read_from_proc (*it);
       match_modules (m, op);
    */
  }
#endif
}

void input_subsystem::parse_modules (const std::string &fname)
{
  std::ifstream f (fname.c_str());
  if (!f.is_open()) return;

  f.setf (std::ios::hex, std::ios::basefield);

  while (!f.eof()) {
    
    // quick hack to skip the leading comment line ...
    if (f.peek() == '#') {
      std::string temp;
      std::getline (f, temp);
      continue;
    }
    
    input_module m;

    f >> m.module >> m.matchbits >> m.bustype >> m.vendor >> m.product >> m.version;

    std::string buf;
    for(int i=0; i<input_module::BITS_LAST; i++) {
      /* FIXME bug in ifstream? */ while ((f.peek() == ' ') || (f.peek() == '\t')) f.get();
      f >> buf;

      /* clear and check */
      m.bits[i] = std::vector<uint32_t>();
    
      const char *ptr = buf.c_str();
      while (ptr && *ptr) {
	char *eptr;
	uint32_t val = strtol(ptr, &eptr, 16);
	
	if (eptr == ptr) break;

	m.bits[i].push_back(val);
	ptr = eptr;
	if (*ptr == ':') ptr++;
      }
    }

    f >> m.driver_info;

    if (f.good()) {
      modules.push_back (m);
    }

    if (f.fail() && !f.eof())
      f.clear (f.rdstate() & ~std::ios::failbit);
  }
}

void input_subsystem::parse_modules ()
{
  if (modules.size())
    return;
  
  parse_modules("/lib/modules/" + kernel_release() + "/modules.inputmap");
  Utility::Glob maps("/etc/hotplug/input.*map");
  for (Utility::Glob::iterator it = maps.begin(); it != maps.end(); ++it)
    parse_modules(*it);
}
