
#include <string>
#include <fstream>
#include <iostream>

#include <Glob.hh>
#include <support.hh>
#include <os.hh>

#include <inttypes.h>

#include <pci-match.hh>

bool pci_module::operator== (const pci_module& dev) {
  const uint32_t any = uint32_t(-1);
  
  if (vendor != any && vendor != dev.vendor)
    return false;
  if (device != any && device != dev.device)
    return false;
  
  if (subvendor != any && subvendor != dev.subvendor)
    return false;
  if (subdevice != any && subdevice != dev.subdevice)
      return false;
  
  if ((dev.pciclass & class_mask) == pciclass)
    return true;
  
  return false;
}

void pci_module::read_from_sysfs (const std::string& file) {
  vendor = read_value_from_file<uint32_t> (file + "/vendor");
  device = read_value_from_file<uint32_t> (file + "/device");
  subvendor = read_value_from_file<uint32_t> (file + "/subsystem_vendor");
  subdevice = read_value_from_file<uint32_t> (file + "/subsystem_device");
  pciclass = read_value_from_file<uint32_t> (file + "/class");
}

void pci_module::read_from_proc (const std::string& file) {
  
  std::ifstream f (file.c_str());
  
  unsigned char cs [64];
  f.read ((char*)cs, sizeof (cs));
  
  vendor = cs[1] << 8 | cs[0];
  device = cs[3] << 8 | cs[2];
  subvendor = cs[0x2c+1] << 8 | cs[0x2c];
  subdevice = cs[0x2c+3] << 8 | cs[0x2c+2];
  
  pciclass = (cs[11] << 16) | (cs[10] << 8) | cs[9];
}

std::ostream& operator<< (std::ostream& os, const pci_module& m) {
  os.setf (std::ios::hex, std::ios::basefield);

  os << "vendor: " << m.vendor
     << ", device. " << m.device
     << ", subvendor: " << m.subvendor
     << ", subdevice: " << m.subdevice
     << ", class: " << m.pciclass
     << ", class mask: " << m.class_mask;
  
  if (!m.module.empty())
    os << ": Module: " << m.module;
  
  return os;
}

void pci_module::read_from_env () {
  // ACTION=%s [add or remove]

    std::string classstr = getenv("PCI_CLASS");
    std::string idstr = getenv("PCI_ID");
    std::string slotstr = getenv("PCI_SLOT_NAME");
    std::string subidstr = getenv("PCI_SUBSYS_ID");
    
    std::string::size_type idx;

    // PCI_CLASS=%06X
    if (!classstr.empty())
      pciclass = strtol(classstr.c_str(), NULL, 16);

    // PCI_ID=%04X:%04X
    if (!idstr.empty()) {
      vendor = strtol(idstr.c_str(), NULL, 16);
      idx = idstr.find(':')+1;
      device = strtol(idstr.c_str()+idx, NULL, 16);
    }
    // PCI_SLOT_NAME=%s
    // FIXME

    // PCI_SUBSYS_ID=%04X:%04X
    if (!subidstr.empty()) {
      subvendor = strtol(subidstr.c_str(), NULL, 16);
      idx = subidstr.find(':')+1;
      subdevice = strtol(subidstr.c_str()+idx, NULL, 16);
    }
}

// ...

std::vector<pci_module> pci_subsystem::modules;

void pci_subsystem::hotplug ()
{
  parse_modules();
  pci_module m;
  m.read_from_env ();
  match_modules (m, &pci_subsystem::hotplug);
}

void pci_subsystem::hotplug (const pci_module& m)
{
  if (m.module.size()) {
    modprobe (m.module);
  }
}

void pci_subsystem::hotplug_all ()
{
  parse_modules();
  for_each (&pci_subsystem::hotplug);
}

void pci_subsystem::list ()
{
  std::cout << "PCI Devices and matched modules:" << std::endl;
  parse_modules();
  
  for_each (&pci_subsystem::list);
  std::cout << std::endl;
}

void pci_subsystem::list (const pci_module& m)
{
  std::cout << m << std::endl;
}

void pci_subsystem::match_modules (pci_module& m,
				   void(pci_subsystem::*op)(const pci_module&))
{
  std::vector<pci_module>::iterator module = modules.begin();
  while (module != modules.end()) {
    module = std::find (module, modules.end(), m);
    if (module != modules.end()) {
      m.module = module->module;
        (this->*op)(m);
	++module;
    }
  }
}

void pci_subsystem::for_each (void(pci_subsystem::*op)(const pci_module&))
{
  Utility::Glob pci_devices ("/sys/bus/pci/devices/????:??:??.?");
  for (Utility::Glob::iterator it = pci_devices.begin();
       it != pci_devices.end(); ++it)
    {
      pci_module m;
      m.read_from_sysfs (*it);
      match_modules (m, op);
    }
  if (pci_devices)
    return;

  Utility::Glob pci_devices2 ("/proc/bus/pci/*/??.?");
  for (Utility::Glob::iterator it = pci_devices2.begin();
       it != pci_devices2.end(); ++it)
  {
    pci_module m;
    m.read_from_proc (*it);
    match_modules (m, op);
  }
}

void pci_subsystem::parse_modules (const std::string &fname)
{
  std::ifstream f (fname.c_str());
  if (!f.is_open()) return;

  f.setf (std::ios::hex, std::ios::basefield);

  while (!f.eof()) {
    
    // quick hack to skip the leading comment line ...
    if (f.peek() == '#') {
      std::string temp;
      std::getline (f, temp);
      continue;
    }
    
    pci_module m;
    
    f >> m.module >> m.vendor >> m.device >> m.subvendor >> m.subdevice
      >> m.pciclass >> m.class_mask >> m.driver_data;
    
    if (f.good())
      modules.push_back (m);
    
    if (f.fail() && !f.eof())
      f.clear (f.rdstate() & ~std::ios::failbit);
  }
}

void pci_subsystem::parse_modules ()
{
  if (modules.size())
    return;
  
  parse_modules("/lib/modules/" + kernel_release() + "/modules.pcimap");
  Utility::Glob maps("/etc/hotplug/pci.*map");
  for (Utility::Glob::iterator it = maps.begin(); it != maps.end(); ++it)
    parse_modules(*it);
}
